@@ -1,4 +1,7 @@
+from __future__ import annotations
+
 import asyncio
+import contextvars
 import os
 import re
 import signal
@@ -7,34 +10,25 @@ import threading
 import time
 from asyncio import (
     AbstractEventLoop,
-    CancelledError,
     Future,
     Task,
     ensure_future,
-    new_event_loop,
-    set_event_loop,
+    get_running_loop,
     sleep,
 )
-from contextlib import contextmanager
+from contextlib import ExitStack, contextmanager
 from subprocess import Popen
 from traceback import format_tb
 from typing import (
     Any,
-    Awaitable,
     Callable,
     Coroutine,
-    Dict,
-    FrozenSet,
     Generator,
     Generic,
     Hashable,
     Iterable,
-    List,
-    Optional,
-    Tuple,
-    Type,
+    Iterator,
     TypeVar,
-    Union,
     cast,
     overload,
 )
@@ -49,7 +43,7 @@ from prompt_toolkit.eventloop import (
     get_traceback_from_context,
     run_in_executor_with_context,
 )
-from prompt_toolkit.eventloop.utils import call_soon_threadsafe, get_event_loop
+from prompt_toolkit.eventloop.utils import call_soon_threadsafe
 from prompt_toolkit.filters import Condition, Filter, FilterOrBool, to_filter
 from prompt_toolkit.formatted_text import AnyFormattedText
 from prompt_toolkit.input.base import Input
@@ -93,12 +87,6 @@ from prompt_toolkit.utils import Event, in_main_thread
 from .current import get_app_session, set_app
 from .run_in_terminal import in_terminal, run_in_terminal
 
-try:
-    import contextvars
-except ImportError:
-    import prompt_toolkit.eventloop.dummy_contextvars as contextvars  # type: ignore
-
-
 __all__ = [
     "Application",
 ]
@@ -195,38 +183,34 @@ class Application(Generic[_AppResult]):
 
     def __init__(
         self,
-        layout: Optional[Layout] = None,
-        style: Optional[BaseStyle] = None,
+        layout: Layout | None = None,
+        style: BaseStyle | None = None,
         include_default_pygments_style: FilterOrBool = True,
-        style_transformation: Optional[StyleTransformation] = None,
-        key_bindings: Optional[KeyBindingsBase] = None,
-        clipboard: Optional[Clipboard] = None,
+        style_transformation: StyleTransformation | None = None,
+        key_bindings: KeyBindingsBase | None = None,
+        clipboard: Clipboard | None = None,
         full_screen: bool = False,
-        color_depth: Union[
-            ColorDepth, Callable[[], Union[ColorDepth, None]], None
-        ] = None,
+        color_depth: (ColorDepth | Callable[[], ColorDepth | None] | None) = None,
         mouse_support: FilterOrBool = False,
-        enable_page_navigation_bindings: Optional[
-            FilterOrBool
-        ] = None,  # Can be None, True or False.
+        enable_page_navigation_bindings: None
+        | (FilterOrBool) = None,  # Can be None, True or False.
         paste_mode: FilterOrBool = False,
         editing_mode: EditingMode = EditingMode.EMACS,
         erase_when_done: bool = False,
         reverse_vi_search_direction: FilterOrBool = False,
-        min_redraw_interval: Union[float, int, None] = None,
-        max_render_postpone_time: Union[float, int, None] = 0.01,
-        refresh_interval: Optional[float] = None,
-        terminal_size_polling_interval: Optional[float] = 0.5,
+        min_redraw_interval: float | int | None = None,
+        max_render_postpone_time: float | int | None = 0.01,
+        refresh_interval: float | None = None,
+        terminal_size_polling_interval: float | None = 0.5,
         cursor: AnyCursorShapeConfig = None,
-        on_reset: Optional["ApplicationEventHandler[_AppResult]"] = None,
-        on_invalidate: Optional["ApplicationEventHandler[_AppResult]"] = None,
-        before_render: Optional["ApplicationEventHandler[_AppResult]"] = None,
-        after_render: Optional["ApplicationEventHandler[_AppResult]"] = None,
+        on_reset: ApplicationEventHandler[_AppResult] | None = None,
+        on_invalidate: ApplicationEventHandler[_AppResult] | None = None,
+        before_render: ApplicationEventHandler[_AppResult] | None = None,
+        after_render: ApplicationEventHandler[_AppResult] | None = None,
         # I/O.
-        input: Optional[Input] = None,
-        output: Optional[Output] = None,
+        input: Input | None = None,
+        output: Output | None = None,
     ) -> None:
-
         # If `enable_page_navigation_bindings` is not specified, enable it in
         # case of full screen applications only. This can be overridden by the user.
         if enable_page_navigation_bindings is None:
@@ -282,12 +266,13 @@ class Application(Generic[_AppResult]):
         self.input = input or session.input
 
         # List of 'extra' functions to execute before a Application.run.
-        self.pre_run_callables: List[Callable[[], None]] = []
+        self.pre_run_callables: list[Callable[[], None]] = []
 
         self._is_running = False
-        self.future: Optional[Future[_AppResult]] = None
-        self.loop: Optional[AbstractEventLoop] = None
-        self.context: Optional[contextvars.Context] = None
+        self.future: Future[_AppResult] | None = None
+        self.loop: AbstractEventLoop | None = None
+        self._loop_thread: threading.Thread | None = None
+        self.context: contextvars.Context | None = None
 
         #: Quoted insert. This flag is set if we go into quoted insert mode.
         self.quoted_insert = False
@@ -332,7 +317,7 @@ class Application(Generic[_AppResult]):
 
         # Invalidate flag. When 'True', a repaint has been scheduled.
         self._invalidated = False
-        self._invalidate_events: List[
+        self._invalidate_events: list[
             Event[object]
         ] = []  # Collection of 'invalidate' Event objects.
         self._last_redraw_time = 0.0  # Unix timestamp of last redraw. Used when
@@ -344,7 +329,7 @@ class Application(Generic[_AppResult]):
         # If `run_in_terminal` was called. This will point to a `Future` what will be
         # set at the point when the previous run finishes.
         self._running_in_terminal = False
-        self._running_in_terminal_f: Optional[Future[None]] = None
+        self._running_in_terminal_f: Future[None] | None = None
 
         # Trigger initialize callback.
         self.reset()
@@ -432,7 +417,7 @@ class Application(Generic[_AppResult]):
 
         self.exit_style = ""
 
-        self.background_tasks: List[Task[None]] = []
+        self._background_tasks: set[Task[None]] = set()
 
         self.renderer.reset()
         self.key_processor.reset()
@@ -459,7 +444,7 @@ class Application(Generic[_AppResult]):
         """
         if not self._is_running:
             # Don't schedule a redraw if we're not running.
-            # Otherwise, `get_event_loop()` in `call_soon_threadsafe` can fail.
+            # Otherwise, `get_running_loop()` in `call_soon_threadsafe` can fail.
             # See: https://github.com/dbcli/mycli/issues/797
             return
 
@@ -612,7 +597,7 @@ class Application(Generic[_AppResult]):
         self._request_absolute_cursor_position()
         self._redraw()
 
-    def _pre_run(self, pre_run: Optional[Callable[[], None]] = None) -> None:
+    def _pre_run(self, pre_run: Callable[[], None] | None = None) -> None:
         """
         Called during `run`.
 
@@ -632,7 +617,7 @@ class Application(Generic[_AppResult]):
 
     async def run_async(
         self,
-        pre_run: Optional[Callable[[], None]] = None,
+        pre_run: Callable[[], None] | None = None,
         set_exception_handler: bool = True,
         handle_sigint: bool = True,
         slow_callback_duration: float = 0.5,
@@ -670,19 +655,15 @@ class Application(Generic[_AppResult]):
             # See: https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1553
             handle_sigint = False
 
-        async def _run_async() -> _AppResult:
-            "Coroutine."
-            loop = get_event_loop()
-            f = loop.create_future()
-            self.future = f  # XXX: make sure to set this before calling '_redraw'.
-            self.loop = loop
-            self.context = contextvars.copy_context()
+        async def _run_async(f: "asyncio.Future[_AppResult]") -> _AppResult:
+            context = contextvars.copy_context()
+            self.context = context
 
             # Counter for cancelling 'flush' timeouts. Every time when a key is
             # pressed, we start a 'flush' timer for flushing our escape key. But
             # when any subsequent input is received, a new timer is started and
             # the current timer will be ignored.
-            flush_task: Optional[asyncio.Task[None]] = None
+            flush_task: asyncio.Task[None] | None = None
 
             # Reset.
             # (`self.future` needs to be set when `pre_run` is called.)
@@ -720,6 +701,16 @@ class Application(Generic[_AppResult]):
                         flush_task.cancel()
                     flush_task = self.create_background_task(auto_flush_input())
 
+            def read_from_input_in_context() -> None:
+                # Ensure that key bindings callbacks are always executed in the
+                # current context. This is important when key bindings are
+                # accessing contextvars. (These callbacks are currently being
+                # called from a different context. Underneath,
+                # `loop.add_reader` is used to register the stdin FD.)
+                # (We copy the context to avoid a `RuntimeError` in case the
+                # context is already active.)
+                context.copy().run(read_from_input)
+
             async def auto_flush_input() -> None:
                 # Flush input after timeout.
                 # (Used for flushing the enter key.)
@@ -739,9 +730,8 @@ class Application(Generic[_AppResult]):
 
             # Enter raw mode, attach input and attach WINCH event handler.
             with self.input.raw_mode(), self.input.attach(
-                read_from_input
+                read_from_input_in_context
             ), attach_winch_signal_handler(self._on_resize):
-
                 # Draw UI.
                 self._request_absolute_cursor_position()
                 self._redraw()
@@ -790,74 +780,121 @@ class Application(Generic[_AppResult]):
                         # Store unprocessed input as typeahead for next time.
                         store_typeahead(self.input, self.key_processor.empty_queue())
 
-                return cast(_AppResult, result)
+                return result
 
-        async def _run_async2() -> _AppResult:
+        @contextmanager
+        def set_loop() -> Iterator[AbstractEventLoop]:
+            loop = get_running_loop()
+            self.loop = loop
+            self._loop_thread = threading.current_thread()
+
+            try:
+                yield loop
+            finally:
+                self.loop = None
+                self._loop_thread = None
+
+        @contextmanager
+        def set_is_running() -> Iterator[None]:
             self._is_running = True
             try:
-                # Make sure to set `_invalidated` to `False` to begin with,
-                # otherwise we're not going to paint anything. This can happen if
-                # this application had run before on a different event loop, and a
-                # paint was scheduled using `call_soon_threadsafe` with
-                # `max_postpone_time`.
-                self._invalidated = False
-
-                loop = get_event_loop()
-
-                if handle_sigint:
-                    loop.add_signal_handler(
-                        signal.SIGINT,
-                        lambda *_: loop.call_soon_threadsafe(
-                            self.key_processor.send_sigint
-                        ),
-                    )
-
-                if set_exception_handler:
-                    previous_exc_handler = loop.get_exception_handler()
-                    loop.set_exception_handler(self._handle_exception)
-
-                # Set slow_callback_duration.
-                original_slow_callback_duration = loop.slow_callback_duration
-                loop.slow_callback_duration = slow_callback_duration
-
-                try:
-                    with set_app(self), self._enable_breakpointhook():
-                        try:
-                            result = await _run_async()
-                        finally:
-                            # Wait for the background tasks to be done. This needs to
-                            # go in the finally! If `_run_async` raises
-                            # `KeyboardInterrupt`, we still want to wait for the
-                            # background tasks.
-                            await self.cancel_and_wait_for_background_tasks()
-
-                            # Also remove the Future again. (This brings the
-                            # application back to its initial state, where it also
-                            # doesn't have a Future.)
-                            self.future = None
-
-                        return result
-                finally:
-                    if set_exception_handler:
-                        loop.set_exception_handler(previous_exc_handler)
-
-                    if handle_sigint:
-                        loop.remove_signal_handler(signal.SIGINT)
-
-                    # Reset slow_callback_duration.
-                    loop.slow_callback_duration = original_slow_callback_duration
-
+                yield
             finally:
-                # Set the `_is_running` flag to `False`. Normally this happened
-                # already in the finally block in `run_async` above, but in
-                # case of exceptions, that's not always the case.
                 self._is_running = False
 
-        return await _run_async2()
+        @contextmanager
+        def set_handle_sigint(loop: AbstractEventLoop) -> Iterator[None]:
+            if handle_sigint:
+                loop.add_signal_handler(
+                    signal.SIGINT,
+                    lambda *_: loop.call_soon_threadsafe(
+                        self.key_processor.send_sigint
+                    ),
+                )
+                try:
+                    yield
+                finally:
+                    loop.remove_signal_handler(signal.SIGINT)
+            else:
+                yield
+
+        @contextmanager
+        def set_exception_handler_ctx(loop: AbstractEventLoop) -> Iterator[None]:
+            if set_exception_handler:
+                previous_exc_handler = loop.get_exception_handler()
+                loop.set_exception_handler(self._handle_exception)
+                try:
+                    yield
+                finally:
+                    loop.set_exception_handler(previous_exc_handler)
+
+            else:
+                yield
+
+        @contextmanager
+        def set_callback_duration(loop: AbstractEventLoop) -> Iterator[None]:
+            # Set slow_callback_duration.
+            original_slow_callback_duration = loop.slow_callback_duration
+            loop.slow_callback_duration = slow_callback_duration
+            try:
+                yield
+            finally:
+                # Reset slow_callback_duration.
+                loop.slow_callback_duration = original_slow_callback_duration
+
+        @contextmanager
+        def create_future(
+            loop: AbstractEventLoop,
+        ) -> Iterator[asyncio.Future[_AppResult]]:
+            f = loop.create_future()
+            self.future = f  # XXX: make sure to set this before calling '_redraw'.
+
+            try:
+                yield f
+            finally:
+                # Also remove the Future again. (This brings the
+                # application back to its initial state, where it also
+                # doesn't have a Future.)
+                self.future = None
+
+        with ExitStack() as stack:
+            stack.enter_context(set_is_running())
+
+            # Make sure to set `_invalidated` to `False` to begin with,
+            # otherwise we're not going to paint anything. This can happen if
+            # this application had run before on a different event loop, and a
+            # paint was scheduled using `call_soon_threadsafe` with
+            # `max_postpone_time`.
+            self._invalidated = False
+
+            loop = stack.enter_context(set_loop())
+
+            stack.enter_context(set_handle_sigint(loop))
+            stack.enter_context(set_exception_handler_ctx(loop))
+            stack.enter_context(set_callback_duration(loop))
+            stack.enter_context(set_app(self))
+            stack.enter_context(self._enable_breakpointhook())
+
+            f = stack.enter_context(create_future(loop))
+
+            try:
+                return await _run_async(f)
+            finally:
+                # Wait for the background tasks to be done. This needs to
+                # go in the finally! If `_run_async` raises
+                # `KeyboardInterrupt`, we still want to wait for the
+                # background tasks.
+                await self.cancel_and_wait_for_background_tasks()
+
+        # The `ExitStack` above is defined in typeshed in a way that it can
+        # swallow exceptions. Without next line, mypy would think that there's
+        # a possibility we don't return here. See:
+        # https://github.com/python/mypy/issues/7726
+        assert False, "unreachable"
 
     def run(
         self,
-        pre_run: Optional[Callable[[], None]] = None,
+        pre_run: Callable[[], None] | None = None,
         set_exception_handler: bool = True,
         handle_sigint: bool = True,
         in_thread: bool = False,
@@ -890,7 +927,7 @@ class Application(Generic[_AppResult]):
         """
         if in_thread:
             result: _AppResult
-            exception: Optional[BaseException] = None
+            exception: BaseException | None = None
 
             def run_in_thread() -> None:
                 nonlocal result, exception
@@ -903,14 +940,6 @@ class Application(Generic[_AppResult]):
                     )
                 except BaseException as e:
                     exception = e
-                finally:
-                    # Make sure to close the event loop in this thread. Running
-                    # the application creates a new loop (because we're in
-                    # another thread), but it doesn't get closed automatically
-                    # (also not by the garbage collector).
-                    loop = get_event_loop()
-                    loop.run_until_complete(loop.shutdown_asyncgens())
-                    loop.close()
 
             thread = threading.Thread(target=run_in_thread)
             thread.start()
@@ -920,30 +949,25 @@ class Application(Generic[_AppResult]):
                 raise exception
             return result
 
-        # We don't create a new event loop by default, because we want to be
-        # sure that when this is called multiple times, each call of `run()`
-        # goes through the same event loop. This way, users can schedule
-        # background-tasks that keep running across multiple prompts.
-        try:
-            loop = get_event_loop()
-        except RuntimeError:
-            # Possibly we are not running in the main thread, where no event
-            # loop is set by default. Or somebody called `asyncio.run()`
-            # before, which closes the existing event loop. We can create a new
-            # loop.
-            loop = new_event_loop()
-            set_event_loop(loop)
-
-        return loop.run_until_complete(
-            self.run_async(
-                pre_run=pre_run,
-                set_exception_handler=set_exception_handler,
-                handle_sigint=handle_sigint,
-            )
+        coro = self.run_async(
+            pre_run=pre_run,
+            set_exception_handler=set_exception_handler,
+            handle_sigint=handle_sigint,
         )
+        try:
+            # See whether a loop was installed already. If so, use that. That's
+            # required for the input hooks to work, they are installed using
+            # `set_event_loop`.
+            loop = asyncio.get_event_loop()
+        except RuntimeError:
+            # No loop installed. Run like usual.
+            return asyncio.run(coro)
+        else:
+            # Use existing loop.
+            return loop.run_until_complete(coro)
 
     def _handle_exception(
-        self, loop: AbstractEventLoop, context: Dict[str, Any]
+        self, loop: AbstractEventLoop, context: dict[str, Any]
     ) -> None:
         """
         Handler for event loop exceptions.
@@ -989,6 +1013,10 @@ class Application(Generic[_AppResult]):
         """
         Breakpointhook which uses PDB, but ensures that the application is
         hidden and input echoing is restored during each debugger dispatch.
+
+        This can be called from any thread. In any case, the application's
+        event loop will be blocked while the PDB input is displayed. The event
+        will continue after leaving the debugger.
         """
         app = self
         # Inline import on purpose. We don't want to import pdb, if not needed.
@@ -997,60 +1025,153 @@ class Application(Generic[_AppResult]):
 
         TraceDispatch = Callable[[FrameType, str, Any], Any]
 
-        class CustomPdb(pdb.Pdb):
-            def trace_dispatch(
-                self, frame: FrameType, event: str, arg: Any
-            ) -> TraceDispatch:
+        @contextmanager
+        def hide_app_from_eventloop_thread() -> Generator[None, None, None]:
+            """Stop application if `__breakpointhook__` is called from within
+            the App's event loop."""
+            # Hide application.
+            app.renderer.erase()
+
+            # Detach input and dispatch to debugger.
+            with app.input.detach():
+                with app.input.cooked_mode():
+                    yield
+
+            # Note: we don't render the application again here, because
+            # there's a good chance that there's a breakpoint on the next
+            # line. This paint/erase cycle would move the PDB prompt back
+            # to the middle of the screen.
+
+        @contextmanager
+        def hide_app_from_other_thread() -> Generator[None, None, None]:
+            """Stop application if `__breakpointhook__` is called from a
+            thread other than the App's event loop."""
+            ready = threading.Event()
+            done = threading.Event()
+
+            async def in_loop() -> None:
+                # from .run_in_terminal import in_terminal
+                # async with in_terminal():
+                #     ready.set()
+                #     await asyncio.get_running_loop().run_in_executor(None, done.wait)
+                #     return
+
                 # Hide application.
                 app.renderer.erase()
 
                 # Detach input and dispatch to debugger.
                 with app.input.detach():
                     with app.input.cooked_mode():
+                        ready.set()
+                        # Here we block the App's event loop thread until the
+                        # debugger resumes. We could have used `with
+                        # run_in_terminal.in_terminal():` like the commented
+                        # code above, but it seems to work better if we
+                        # completely stop the main event loop while debugging.
+                        done.wait()
+
+            self.create_background_task(in_loop())
+            ready.wait()
+            try:
+                yield
+            finally:
+                done.set()
+
+        class CustomPdb(pdb.Pdb):
+            def trace_dispatch(
+                self, frame: FrameType, event: str, arg: Any
+            ) -> TraceDispatch:
+                if app._loop_thread is None:
+                    return super().trace_dispatch(frame, event, arg)
+
+                if app._loop_thread == threading.current_thread():
+                    with hide_app_from_eventloop_thread():
                         return super().trace_dispatch(frame, event, arg)
 
-                # Note: we don't render the application again here, because
-                # there's a good chance that there's a breakpoint on the next
-                # line. This paint/erase cycle would move the PDB prompt back
-                # to the middle of the screen.
+                with hide_app_from_other_thread():
+                    return super().trace_dispatch(frame, event, arg)
 
         frame = sys._getframe().f_back
         CustomPdb(stdout=sys.__stdout__).set_trace(frame)
 
     def create_background_task(
         self, coroutine: Coroutine[Any, Any, None]
-    ) -> "asyncio.Task[None]":
+    ) -> asyncio.Task[None]:
         """
         Start a background task (coroutine) for the running application. When
         the `Application` terminates, unfinished background tasks will be
         cancelled.
 
-        If asyncio had nurseries like Trio, we would create a nursery in
-        `Application.run_async`, and run the given coroutine in that nursery.
+        Given that we still support Python versions before 3.11, we can't use
+        task groups (and exception groups), because of that, these background
+        tasks are not allowed to raise exceptions. If they do, we'll call the
+        default exception handler from the event loop.
 
-        Not threadsafe.
+        If at some point, we have Python 3.11 as the minimum supported Python
+        version, then we can use a `TaskGroup` (with the lifetime of
+        `Application.run_async()`, and run run the background tasks in there.
+
+        This is not threadsafe.
         """
-        task: asyncio.Task[None] = get_event_loop().create_task(coroutine)
-        self.background_tasks.append(task)
+        loop = self.loop or get_running_loop()
+        task: asyncio.Task[None] = loop.create_task(coroutine)
+        self._background_tasks.add(task)
+
+        task.add_done_callback(self._on_background_task_done)
         return task
 
+    def _on_background_task_done(self, task: asyncio.Task[None]) -> None:
+        """
+        Called when a background task completes. Remove it from
+        `_background_tasks`, and handle exceptions if any.
+        """
+        self._background_tasks.discard(task)
+
+        if task.cancelled():
+            return
+
+        exc = task.exception()
+        if exc is not None:
+            get_running_loop().call_exception_handler(
+                {
+                    "message": f"prompt_toolkit.Application background task {task!r} "
+                    "raised an unexpected exception.",
+                    "exception": exc,
+                    "task": task,
+                }
+            )
+
     async def cancel_and_wait_for_background_tasks(self) -> None:
         """
-        Cancel all background tasks, and wait for the cancellation to be done.
+        Cancel all background tasks, and wait for the cancellation to complete.
         If any of the background tasks raised an exception, this will also
         propagate the exception.
 
         (If we had nurseries like Trio, this would be the `__aexit__` of a
         nursery.)
         """
-        for task in self.background_tasks:
+        for task in self._background_tasks:
             task.cancel()
 
-        for task in self.background_tasks:
-            try:
-                await task
-            except CancelledError:
-                pass
+        # Wait until the cancellation of the background tasks completes.
+        # `asyncio.wait()` does not propagate exceptions raised within any of
+        # these tasks, which is what we want. Otherwise, we can't distinguish
+        # between a `CancelledError` raised in this task because it got
+        # cancelled, and a `CancelledError` raised on this `await` checkpoint,
+        # because *we* got cancelled during the teardown of the application.
+        # (If we get cancelled here, then it's important to not suppress the
+        # `CancelledError`, and have it propagate.)
+        # NOTE: Currently, if we get cancelled at this point then we can't wait
+        #       for the cancellation to complete (in the future, we should be
+        #       using anyio or Python's 3.11 TaskGroup.)
+        #       Also, if we had exception groups, we could propagate an
+        #       `ExceptionGroup` if something went wrong here. Right now, we
+        #       don't propagate exceptions, but have them printed in
+        #       `_on_background_task_done`.
+        if len(self._background_tasks) > 0:
+            await asyncio.wait(
+                self._background_tasks, timeout=None, return_when=asyncio.ALL_COMPLETED
+            )
 
     async def _poll_output_size(self) -> None:
         """
@@ -1060,7 +1181,7 @@ class Application(Generic[_AppResult]):
         - If we are not running in the main thread.
         - On Windows.
         """
-        size: Optional[Size] = None
+        size: Size | None = None
         interval = self.terminal_size_polling_interval
 
         if interval is None:
@@ -1099,14 +1220,14 @@ class Application(Generic[_AppResult]):
 
     @overload
     def exit(
-        self, *, exception: Union[BaseException, Type[BaseException]], style: str = ""
+        self, *, exception: BaseException | type[BaseException], style: str = ""
     ) -> None:
         "Exit with exception."
 
     def exit(
         self,
-        result: Optional[_AppResult] = None,
-        exception: Optional[Union[BaseException, Type[BaseException]]] = None,
+        result: _AppResult | None = None,
+        exception: BaseException | type[BaseException] | None = None,
         style: str = "",
     ) -> None:
         """
@@ -1221,7 +1342,7 @@ class Application(Generic[_AppResult]):
             run_in_terminal(run)
 
     def print_text(
-        self, text: AnyFormattedText, style: Optional[BaseStyle] = None
+        self, text: AnyFormattedText, style: BaseStyle | None = None
     ) -> None:
         """
         Print a list of (style_str, text) tuples to the output.
@@ -1250,7 +1371,7 @@ class Application(Generic[_AppResult]):
             return self.future.done()
         return False
 
-    def get_used_style_strings(self) -> List[str]:
+    def get_used_style_strings(self) -> list[str]:
         """
         Return a list of used style strings. This is helpful for debugging, and
         for writing a new `Style`.
@@ -1276,7 +1397,7 @@ class _CombinedRegistry(KeyBindingsBase):
     def __init__(self, app: Application[_AppResult]) -> None:
         self.app = app
         self._cache: SimpleCache[
-            Tuple[Window, FrozenSet[UIControl]], KeyBindingsBase
+            tuple[Window, frozenset[UIControl]], KeyBindingsBase
         ] = SimpleCache()
 
     @property
@@ -1285,13 +1406,14 @@ class _CombinedRegistry(KeyBindingsBase):
         KeyBindings object."""
         raise NotImplementedError
 
-    def bindings(self) -> List[Binding]:
+    @property
+    def bindings(self) -> list[Binding]:
         """Not needed - this object is not going to be wrapped in another
         KeyBindings object."""
         raise NotImplementedError
 
     def _create_key_bindings(
-        self, current_window: Window, other_controls: List[UIControl]
+        self, current_window: Window, other_controls: list[UIControl]
     ) -> KeyBindingsBase:
         """
         Create a `KeyBindings` object that merges the `KeyBindings` from the
@@ -1354,10 +1476,10 @@ class _CombinedRegistry(KeyBindingsBase):
             key, lambda: self._create_key_bindings(current_window, other_controls)
         )
 
-    def get_bindings_for_keys(self, keys: KeysTuple) -> List[Binding]:
+    def get_bindings_for_keys(self, keys: KeysTuple) -> list[Binding]:
         return self._key_bindings.get_bindings_for_keys(keys)
 
-    def get_bindings_starting_with_keys(self, keys: KeysTuple) -> List[Binding]:
+    def get_bindings_starting_with_keys(self, keys: KeysTuple) -> list[Binding]:
         return self._key_bindings.get_bindings_starting_with_keys(keys)
 
 
@@ -1384,7 +1506,10 @@ async def _do_wait_for_enter(wait_text: AnyFormattedText) -> None:
     session: PromptSession[None] = PromptSession(
         message=wait_text, key_bindings=key_bindings
     )
-    await session.app.run_async()
+    try:
+        await session.app.run_async()
+    except KeyboardInterrupt:
+        pass  # Control-c pressed. Don't propagate this error.
 
 
 @contextmanager
@@ -1415,7 +1540,7 @@ def attach_winch_signal_handler(
 
     # Keep track of the previous handler.
     # (Only UnixSelectorEventloop has `_signal_handlers`.)
-    loop = get_event_loop()
+    loop = get_running_loop()
     previous_winch_handler = getattr(loop, "_signal_handlers", {}).get(sigwinch)
 
     try:
