@@ -1,17 +1,9 @@
+from __future__ import annotations
+
 import math
 from itertools import zip_longest
-from typing import (
-    TYPE_CHECKING,
-    Callable,
-    Dict,
-    Iterable,
-    List,
-    Optional,
-    Tuple,
-    TypeVar,
-    Union,
-    cast,
-)
+from typing import TYPE_CHECKING, Callable, Iterable, Sequence, Tuple, TypeVar, cast
+from weakref import WeakKeyDictionary
 
 from prompt_toolkit.application.current import get_app
 from prompt_toolkit.buffer import CompletionState
@@ -40,9 +32,11 @@ from .dimension import Dimension
 from .margins import ScrollbarMargin
 
 if TYPE_CHECKING:
-    from prompt_toolkit.key_binding.key_bindings import KeyBindings
+    from prompt_toolkit.key_binding.key_bindings import (
+        KeyBindings,
+        NotImplementedOrNone,
+    )
 
-    NotImplementedOrNone = object
 
 __all__ = [
     "CompletionsMenu",
@@ -70,7 +64,7 @@ class CompletionsMenuControl(UIControl):
     def has_focus(self) -> bool:
         return False
 
-    def preferred_width(self, max_available_width: int) -> Optional[int]:
+    def preferred_width(self, max_available_width: int) -> int | None:
         complete_state = get_app().current_buffer.complete_state
         if complete_state:
             menu_width = self._get_menu_width(500, complete_state)
@@ -85,9 +79,8 @@ class CompletionsMenuControl(UIControl):
         width: int,
         max_available_height: int,
         wrap_lines: bool,
-        get_line_prefix: Optional[GetLinePrefixCallable],
-    ) -> Optional[int]:
-
+        get_line_prefix: GetLinePrefixCallable | None,
+    ) -> int | None:
         complete_state = get_app().current_buffer.complete_state
         if complete_state:
             return len(complete_state.completions)
@@ -160,16 +153,19 @@ class CompletionsMenuControl(UIControl):
             return get_cwidth(completion.display_meta_text)
 
         if self._show_meta(complete_state):
-            return min(
-                max_width, max(meta_width(c) for c in complete_state.completions) + 2
-            )
+            # If the amount of completions is over 200, compute the width based
+            # on the first 200 completions, otherwise this can be very slow.
+            completions = complete_state.completions
+            if len(completions) > 200:
+                completions = completions[:200]
+
+            return min(max_width, max(meta_width(c) for c in completions) + 2)
         else:
             return 0
 
     def _get_menu_item_meta_fragments(
         self, completion: Completion, is_current_completion: bool, width: int
     ) -> StyleAndTextTuples:
-
         if is_current_completion:
             style_str = "class:completion-menu.meta.completion.current"
         else:
@@ -183,7 +179,7 @@ class CompletionsMenuControl(UIControl):
             style=style_str,
         )
 
-    def mouse_handler(self, mouse_event: MouseEvent) -> "NotImplementedOrNone":
+    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
         """
         Handle mouse events: clicking and scrolling.
         """
@@ -216,7 +212,7 @@ def _get_menu_item_fragments(
     width.
     """
     if is_current_completion:
-        style_str = "class:completion-menu.completion.current %s %s" % (
+        style_str = "class:completion-menu.completion.current {} {}".format(
             completion.style,
             completion.selected_style,
         )
@@ -237,7 +233,7 @@ def _get_menu_item_fragments(
 
 def _trim_formatted_text(
     formatted_text: StyleAndTextTuples, max_width: int
-) -> Tuple[StyleAndTextTuples, int]:
+) -> tuple[StyleAndTextTuples, int]:
     """
     Trim the text to `max_width`, append dots when the text is too long.
     Returns (text, width) tuple.
@@ -271,13 +267,12 @@ class CompletionsMenu(ConditionalContainer):
     #       visible at the point where we draw this menu.
     def __init__(
         self,
-        max_height: Optional[int] = None,
-        scroll_offset: Union[int, Callable[[], int]] = 0,
+        max_height: int | None = None,
+        scroll_offset: int | Callable[[], int] = 0,
         extra_filter: FilterOrBool = True,
         display_arrows: FilterOrBool = False,
-        z_index: int = 10 ** 8,
+        z_index: int = 10**8,
     ) -> None:
-
         extra_filter = to_filter(extra_filter)
         display_arrows = to_filter(display_arrows)
 
@@ -294,7 +289,7 @@ class CompletionsMenu(ConditionalContainer):
             ),
             # Show when there are completions but not at the point we are
             # returning the input.
-            filter=has_completions & ~is_done & extra_filter,
+            filter=extra_filter & has_completions & ~is_done,
         )
 
 
@@ -330,11 +325,21 @@ class MultiColumnCompletionMenuControl(UIControl):
         self.suggested_max_column_width = suggested_max_column_width
         self.scroll = 0
 
+        # Cache for column width computations. This computation is not cheap,
+        # so we don't want to do it over and over again while the user
+        # navigates through the completions.
+        # (map `completion_state` to `(completion_count, width)`. We remember
+        # the count, because a completer can add new completions to the
+        # `CompletionState` while loading.)
+        self._column_width_for_completion_state: WeakKeyDictionary[
+            CompletionState, Tuple[int, int]
+        ] = WeakKeyDictionary()
+
         # Info of last rendering.
         self._rendered_rows = 0
         self._rendered_columns = 0
         self._total_columns = 0
-        self._render_pos_to_completion: Dict[Tuple[int, int], Completion] = {}
+        self._render_pos_to_completion: dict[tuple[int, int], Completion] = {}
         self._render_left_arrow = False
         self._render_right_arrow = False
         self._render_width = 0
@@ -345,7 +350,7 @@ class MultiColumnCompletionMenuControl(UIControl):
     def has_focus(self) -> bool:
         return False
 
-    def preferred_width(self, max_available_width: int) -> Optional[int]:
+    def preferred_width(self, max_available_width: int) -> int | None:
         """
         Preferred width: prefer to use at least min_rows, but otherwise as much
         as possible horizontally.
@@ -375,8 +380,8 @@ class MultiColumnCompletionMenuControl(UIControl):
         width: int,
         max_available_height: int,
         wrap_lines: bool,
-        get_line_prefix: Optional[GetLinePrefixCallable],
-    ) -> Optional[int]:
+        get_line_prefix: GetLinePrefixCallable | None,
+    ) -> int | None:
         """
         Preferred height: as much as needed in order to display all the completions.
         """
@@ -403,8 +408,8 @@ class MultiColumnCompletionMenuControl(UIControl):
         _T = TypeVar("_T")
 
         def grouper(
-            n: int, iterable: Iterable[_T], fillvalue: Optional[_T] = None
-        ) -> Iterable[List[_T]]:
+            n: int, iterable: Iterable[_T], fillvalue: _T | None = None
+        ) -> Iterable[Sequence[_T | None]]:
             "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
             args = [iter(iterable)] * n
             return zip_longest(fillvalue=fillvalue, *args)
@@ -506,13 +511,28 @@ class MultiColumnCompletionMenuControl(UIControl):
 
         return UIContent(get_line=get_line, line_count=len(rows_))
 
-    def _get_column_width(self, complete_state: CompletionState) -> int:
+    def _get_column_width(self, completion_state: CompletionState) -> int:
         """
         Return the width of each column.
         """
-        return max(get_cwidth(c.display_text) for c in complete_state.completions) + 1
+        try:
+            count, width = self._column_width_for_completion_state[completion_state]
+            if count != len(completion_state.completions):
+                # Number of completions changed, recompute.
+                raise KeyError
+            return width
+        except KeyError:
+            result = (
+                max(get_cwidth(c.display_text) for c in completion_state.completions)
+                + 1
+            )
+            self._column_width_for_completion_state[completion_state] = (
+                len(completion_state.completions),
+                result,
+            )
+            return result
 
-    def mouse_handler(self, mouse_event: MouseEvent) -> "NotImplementedOrNone":
+    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
         """
         Handle scroll and click events.
         """
@@ -556,7 +576,7 @@ class MultiColumnCompletionMenuControl(UIControl):
 
         return None
 
-    def get_key_bindings(self) -> "KeyBindings":
+    def get_key_bindings(self) -> KeyBindings:
         """
         Expose key bindings that handle the left/right arrow keys when the menu
         is displayed.
@@ -620,15 +640,14 @@ class MultiColumnCompletionsMenu(HSplit):
         suggested_max_column_width: int = 30,
         show_meta: FilterOrBool = True,
         extra_filter: FilterOrBool = True,
-        z_index: int = 10 ** 8,
+        z_index: int = 10**8,
     ) -> None:
-
         show_meta = to_filter(show_meta)
         extra_filter = to_filter(extra_filter)
 
         # Display filter: show when there are completions but not at the point
         # we are returning the input.
-        full_filter = has_completions & ~is_done & extra_filter
+        full_filter = extra_filter & has_completions & ~is_done
 
         @Condition
         def any_completion_has_meta() -> bool:
@@ -657,7 +676,7 @@ class MultiColumnCompletionsMenu(HSplit):
 
         meta_window = ConditionalContainer(
             content=Window(content=_SelectedCompletionMetaControl()),
-            filter=show_meta & full_filter & any_completion_has_meta,
+            filter=full_filter & show_meta & any_completion_has_meta,
         )
 
         # Initialise split.
@@ -669,7 +688,7 @@ class _SelectedCompletionMetaControl(UIControl):
     Control that shows the meta information of the selected completion.
     """
 
-    def preferred_width(self, max_available_width: int) -> Optional[int]:
+    def preferred_width(self, max_available_width: int) -> int | None:
         """
         Report the width of the longest meta text as the preferred width of this control.
 
@@ -680,7 +699,19 @@ class _SelectedCompletionMetaControl(UIControl):
         app = get_app()
         if app.current_buffer.complete_state:
             state = app.current_buffer.complete_state
-            return 2 + max(get_cwidth(c.display_meta_text) for c in state.completions)
+
+            if len(state.completions) >= 30:
+                # When there are many completions, calling `get_cwidth` for
+                # every `display_meta_text` is too expensive. In this case,
+                # just return the max available width. There will be enough
+                # columns anyway so that the whole screen is filled with
+                # completions and `create_content` will then take up as much
+                # space as needed.
+                return max_available_width
+
+            return 2 + max(
+                get_cwidth(c.display_meta_text) for c in state.completions[:100]
+            )
         else:
             return 0
 
@@ -689,8 +720,8 @@ class _SelectedCompletionMetaControl(UIControl):
         width: int,
         max_available_height: int,
         wrap_lines: bool,
-        get_line_prefix: Optional[GetLinePrefixCallable],
-    ) -> Optional[int]:
+        get_line_prefix: GetLinePrefixCallable | None,
+    ) -> int | None:
         return 1
 
     def create_content(self, width: int, height: int) -> UIContent:
